/*=========================================================================
| (c) 2017  Total Phase, Inc.
|--------------------------------------------------------------------------
| Project ) PD Analyzer Sample Code
| File    ) capture_usbpd.c
|--------------------------------------------------------------------------
| Simple Capture Example
|--------------------------------------------------------------------------
| Redistribution and use of this file in source and binary forms, with
| or without modification, are permitted.
|
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
| "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
| LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
| FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
| COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
| INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
| BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
| LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
| ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
| POSSIBILITY OF SUCH DAMAGE.
 ========================================================================*/

/*=========================================================================
| INCLUDES
 ========================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pd.h"


/*=========================================================================
| MACROS
 ========================================================================*/
/* The printf format strings are differenct for 64-bit integers
   between Windows and Linux */
#if defined(WIN32) || defined(_WIN32)
#  define U64_FORMAT_STR "I64u"
#else
#  include <inttypes.h>
#  define U64_FORMAT_STR PRIu64
#endif


/*=========================================================================
| PRINT FUNCTIONS
 ========================================================================*/
int print_pd_info (int idx, PdReadInfo *info, const char *info_type)
{
    int cc, pos_err;

    printf("%d, %" U64_FORMAT_STR ", %s, ", idx, info->timestamp, info_type);

    if (strcmp(info_type, "PD"))
        return 1;

    cc = (info->events & PD_EVENT_USBPD_CC_MASK) >> PD_EVENT_USBPD_CC_SHIFT;
    printf("%d, ", (cc + 1));

    if (info->events & PD_EVENT_USBPD_POL_CHANGE) {
        printf("POL_CHANGE\n");
        return 0;
    }
    else if ((info->events & PD_EVENT_USBPD_SOP_MASK) ==
             PD_EVENT_USBPD_HARD_RESET) {
        printf("HARD_RESET\n");
        return 0;
    }
    else if ((info->events & PD_EVENT_USBPD_SOP_MASK) ==
             PD_EVENT_USBPD_CABLE_RESET) {
        printf("CABLE_RESET\n");
        return 0;
    }

    pos_err = info->status & PD_STATUS_USBPD_ERR_MASK;
    if (pos_err == PD_STATUS_USBPD_ERR_PREAMBLE)
        printf("ERR_PREAMBLE, ");
    else if (pos_err == PD_STATUS_USBPD_ERR_SOP)
        printf("ERR_SOP, ");
    else if (pos_err == PD_STATUS_USBPD_ERR_HEADER)
        printf("ERR_HEADER, ");
    else if (pos_err == PD_STATUS_USBPD_ERR_DATA)
        printf("ERR_DATA, ");
    else if (pos_err == PD_STATUS_USBPD_ERR_CRC)
        printf("ERR_CRC, ");
    else if (pos_err == PD_STATUS_USBPD_ERR_EOP)
        printf("ERR_EOP, ");

    if (info->status & PD_STATUS_USBPD_ERR_BAD_CRC)
        printf("BAD_CRC, ");
    else if (pos_err == 0)
        printf("OK, ");

    return 1;
}

void print_bits (int idx, PdReadInfo *info, u08 *bits, u32 pl, u32 bl)
{
    const char *kcode_str[] = {
        "Error", "Error", "Error", "Error", "Error", "Error", "SYNC3", "RST1",
        "Error", "1", "4", "5", "Error", "EOP", "6", "7",
        "Error", "SYNC2", "8", "9", "2", "3", "A", "B",
        "SYNC1", "RST2", "C", "D", "E", "F", "0", "Error"
    };
    int ii, jj, kcode, off, bit;

    if (!print_pd_info(idx, info, "PD"))
        return;

    printf("%d, ", pl);

    for (ii = pl; ii < bl; ii += 5) {
        if (ii + 5 > bl)
            break;

        kcode = 0;
        for (jj = 0; jj < 5; ++jj) {
            off = ii + jj;
            bit = (bits[off / 8] >> (7 - (off % 8))) & 0x1;

            kcode |= (bit << jj);
        }

        printf("%s ", kcode_str[kcode]);
    }

    printf("\n");
}

void print_pd_packets (int idx, PdReadInfo *info,
                       u32 preamble, u32 header, u32 crc,
                       int len, u08 *data)
{
    const char *sop_str[] = {
        "SOP", "SOP\'", "SOP\"", "SOP\'D", "SOP\"D", "", "", ""
    };
    int pos_err, sop, offset;

    if (!print_pd_info(idx, info, "PD"))
        return;

    pos_err = info->status & PD_STATUS_USBPD_ERR_MASK;

    if (pos_err == PD_STATUS_USBPD_ERR_PREAMBLE ||
        pos_err == PD_STATUS_USBPD_ERR_SOP) {
        printf("\n");
        return;
    }

    sop = info->events & PD_EVENT_USBPD_SOP_MASK;

    printf("%s, ", sop_str[sop]);

    if (pos_err == PD_STATUS_USBPD_ERR_HEADER) {
        printf("\n");
        return;
    }

    if (info->events & PD_EVENT_USBPD_EXTENDED_HEADER) {
        printf("%04x;%04x, ",
               header & PD_HEADER_USBPD_STANDARD_MASK,
               (header & PD_HEADER_USBPD_EXTENDED_MASK) >> 16);
    }
    else {
        printf("%04x, ", header & PD_HEADER_USBPD_STANDARD_MASK);
    }

    if (len > 0) {
        for (offset = 0; offset < len; ++offset) {
            printf("%02x ", data[offset]);
        }
    }

    printf(", ");

    if (pos_err == PD_STATUS_USBPD_ERR_DATA ||
        pos_err == PD_STATUS_USBPD_ERR_CRC) {
        printf("\n");
        return;
    }

    if (info->status & PD_STATUS_USBPD_ERR_BAD_CRC)
        printf("%08x(NOT OK)\n", crc);
    else
        printf("%08x(OK)\n", crc);
}

void print_pd_iv (int idx, PdReadInfo *info, int val)
{
    const char *iv_type_str[] = {
        "Vbus Voltage", "Vbus Current", "Vconn Voltage", "Vconn Current",
        "CC1 Voltage", "CC1 Current", "CC2 Voltage", "CC2 Current",
    };
    int iv_type;

    if (!print_pd_info(idx, info, "PD_IV"))
        return;

    iv_type = ((info->events & PD_EVENT_USBPD_IV_SOURCE_MASK)
               >> PD_EVENT_USBPD_IV_SHIFT);
    printf("%s, %d%s\n",
           iv_type_str[iv_type], val, (iv_type & 1)? "mA": "mV");
}


/*=========================================================================
| DUMP FUNCTIONS
 ========================================================================*/
static void dump_pd_raw (Pd pd, int num_packets)
{
    int packetnum = 0;
    PdReadInfo info;
    int ret;
    u32 bl, pl;
    u08 bits[512];

    printf("index,time(us),PD,status,cc,preamble,bits\n");

    while (num_packets <= 0 || packetnum < num_packets) {
        // maximum bit length of USB PD v3.0 is 2769 bits = 347 bytes
        ret = pd_usbpd_read_bits(pd, &info, &bl, &pl, sizeof(bits), bits);
        if (ret == PD_READ_EMPTY)
            continue;

        if (ret < 0) {
            printf("Failed to capture PD on port) %d(%s)\n",
                   ret, pd_status_string(ret));
            break;
        }

        print_bits(packetnum, &info, bits, pl, bl);

        packetnum += 1;
    }
}

static void dump_pd_data (Pd pd, int num_packets)
{
    int packetnum = 0;
    PdReadInfo info;
    int ret;
    u32 pl, header, crc;
    u08 data[512];

    printf("index,time(us),PD,cc,status,sop,header,data0 ... dataN,crc\n");

    while (num_packets <= 0 || packetnum < num_packets) {
        // maximum bit length of USB PD v3.0 is 2769 bits = 347 bytes
        ret = pd_usbpd_read_data(pd, &info, &pl, &header, &crc,
                                 sizeof(data), data);
        if (ret == PD_READ_EMPTY)
            continue;

        if (ret < 0) {
            printf("Failed to capture PD on port) %d(%s)\n",
                   ret, pd_status_string(ret));
            break;
        }

        print_pd_packets(packetnum, &info, pl, header, crc, ret, data);

        packetnum += 1;
    }
}

static void dump_pd_iv (Pd pd, int num_packets)
{
    int packetnum = 0;
    PdReadInfo info;
    int ret, val;

    printf("index,time(us),PD_IV,type,value\n");

    while (num_packets <= 0 || packetnum < num_packets) {
        ret = pd_usbpd_read_iv(pd, &info, &val);
        if (ret == PD_READ_EMPTY)
            continue;

        if (ret < 0) {
            printf("Failed to capture PD on port) %d(%s)\n",
                   ret, pd_status_string(ret));
            break;
        }

        print_pd_iv(packetnum, &info, val);

        packetnum += 1;
    }
}

static void dump_pd_all (Pd pd, int num_packets)
{
    int packetnum = 0;
    PdReadInfo info;
    int ret, val;
    u32 pl, header, crc;
    u08 data[512];

    printf("index,time(us),PD,cc,status,sop,header,data0 ... dataN,crc\n"
           "or\n"
           "index,time(us),PD_IV,type,value\n");

    while (num_packets <= 0 || packetnum < num_packets) {
        // maximum bit length of USB PD v3.0 is 2769 bits = 347 bytes
        ret = pd_usbpd_read_data(pd, &info, &pl, &header, &crc,
                                 sizeof(data), data);
        if (ret != PD_READ_EMPTY) {
            if (ret < 0) {
                printf("Failed to capture PD on port) %d(%s)\n",
                       ret, pd_status_string(ret));
                break;
            }

            print_pd_packets(packetnum, &info, pl, header, crc, ret, data);

            packetnum += 1;
        }

        ret = pd_usbpd_read_iv(pd, &info, &val);
        if (ret != PD_READ_EMPTY) {
            if (ret < 0) {
                printf("Failed to capture PD on port) %d(%s)\n",
                       ret, pd_status_string(ret));
                break;
            }

            print_pd_iv(packetnum, &info, val);

            packetnum += 1;
        }
    }
}


/*=========================================================================
| USAGE INFORMATION
 ========================================================================*/
static void print_usage (void)
{
    printf(
"Usage) capture_usbpd port mode num_events\n"
"Example utility for capturing USB PD data from Power Delivery analyzers.\n"
"\n"
"  The parameter mode can be one of 'raw', 'data', 'iv', and 'all'.\n"
"    'raw' mode shows raw data after bmc decoding, 'data' mode shows pd packets,\n"
"    'iv' mode shows voltage and current, and 'all' is combined of 'data' and 'iv'\n"
"\n"
"  The parameter num_events is set to the number of events to process\n"
"    before exiting.  If num_events is set to zero, the capture will continue\n"
"    indefinitely.\n"
"\n"
"For product documentation and specifications, see www.totalphase.com.\n");
    fflush(stdout);
};


/*=========================================================================
| MAIN PROGRAM ENTRY POINT
 ========================================================================*/
int main (int argc, char *argv[])
{
    Pd pd            = 0;
    int port         = 0;      // open port 0 by default
    const char *mode = NULL;
    int num          = 0;

    if (argc < 4) {
        print_usage();
        exit(1);
    }

    port = atoi(argv[1]);
    mode = argv[2];
    num  = atoi(argv[3]);

    // Open the device
    pd = pd_open(port);
    if (pd <= 0) {
        printf("Unable to open PD Analyzer on port %d\n", port);
        printf("Error code = %d (%s)\n", pd, pd_status_string(pd));
        return 1;
    }
    printf("Opened PD Analyzer on port %d\n", port);

    // Start capture
    pd_capture_start(pd);

    printf("Start capturing\n");
    if (strcmp(mode, "raw") == 0)
        dump_pd_raw(pd, num);
    if (strcmp(mode, "data") == 0)
        dump_pd_data(pd, num);
    if (strcmp(mode, "iv") == 0)
        dump_pd_iv(pd, num);
    if (strcmp(mode, "all") == 0)
        dump_pd_all(pd, num);

    // Stop capture
    pd_capture_stop(pd);

    // Close the device
    pd_close(pd);

    return 0;
}
