/*=========================================================================
| Power Delivery Analyzer Library
|--------------------------------------------------------------------------
| Copyright (c) 2016-2024 Total Phase, Inc.
| All rights reserved.
| www.totalphase.com
|
| Redistribution and use of this file in source and binary forms, with
| or without modification, are permitted provided that the following
| conditions are met:
|
| - Redistributions of source code must retain the above copyright
|   notice, this list of conditions, and the following disclaimer.
|
| - Redistributions in binary form must reproduce the above copyright
|   notice, this list of conditions, and the following disclaimer in the
|   documentation or other materials provided with the distribution.
|
| - This file must only be used to interface with Total Phase products.
|   The names of Total Phase and its contributors must not be used to
|   endorse or promote products derived from this software.
|
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
| "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING BUT NOT
| LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
| FOR A PARTICULAR PURPOSE, ARE DISCLAIMED.  IN NO EVENT WILL THE
| COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
| INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING
| BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
| LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
| ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
| POSSIBILITY OF SUCH DAMAGE.
|--------------------------------------------------------------------------
| To access Total Phase Power Delivery analyzers through the API:
|
| 1) Use one of the following shared objects:
|      pd.so       --  Linux or macOS shared object
|      pd.dll      --  Windows dynamic link library
|
| 2) Along with one of the following language modules:
|      pd.c/h      --  C/C++ API header file and interface module
|      pd_py.py    --  Python API
|      pd.cs       --  C# .NET source
|      pd_net.dll  --  Compiled .NET binding
 ========================================================================*/


/*=========================================================================
| INCLUDES
 ========================================================================*/
/* This #include can be customized to conform to the user's build paths. */
#include "pd.h"


/*=========================================================================
| VERSION CHECK
 ========================================================================*/
#define PD_CFILE_VERSION   0x0200   /* v2.00 */
#define PD_REQ_SW_VERSION  0x0128   /* v1.40 */

/*
 * Make sure that the header file was included and that
 * the version numbers match.
 */
#ifndef PD_HEADER_VERSION
#  error Unable to include header file. Please check include path.

#elif PD_HEADER_VERSION != PD_CFILE_VERSION
#  error Version mismatch between source and header files.

#endif


/*=========================================================================
| DEFINES
 ========================================================================*/
#define API_NAME                     "pd"
#define API_DEBUG                    PD_DEBUG
#define API_OK                       PD_OK
#define API_UNABLE_TO_LOAD_LIBRARY   PD_UNABLE_TO_LOAD_LIBRARY
#define API_INCOMPATIBLE_LIBRARY     PD_INCOMPATIBLE_LIBRARY
#define API_UNABLE_TO_LOAD_FUNCTION  PD_UNABLE_TO_LOAD_FUNCTION
#define API_HEADER_VERSION           PD_HEADER_VERSION
#define API_REQ_SW_VERSION           PD_REQ_SW_VERSION


/*=========================================================================
| LINUX AND DARWIN SUPPORT
 ========================================================================*/
#if defined(__APPLE_CC__) && !defined(DARWIN)
#define DARWIN
#endif

#if defined(linux) || defined(DARWIN)

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>

#ifdef DARWIN
#define DLOPEN_NO_WARN
extern int _NSGetExecutablePath (char *buf, unsigned long *bufsize);
#endif

#include <dlfcn.h>

#define DLL_HANDLE  void *
#define MAX_SO_PATH 256

static char SO_NAME[MAX_SO_PATH+1] = API_NAME ".so";

/*
 * These functions allow the Linux behavior to emulate
 * the Windows behavior as specified below in the Windows
 * support section.
 *
 * First, search for the shared object in the application
 * binary path, then in the current working directory.
 *
 * Searching the application binary path requires /proc
 * filesystem support, which is standard in 2.4.x kernels.
 *
 * If the /proc filesystem is not present, the shared object
 * will not be loaded from the execution path unless that path
 * is either the current working directory or explicitly
 * specified in LD_LIBRARY_PATH.
 */
static int _checkPath (const char *path) {
    char *filename = (char *)malloc(strlen(path) +1 + strlen(SO_NAME) +1);
    int   fd;

    // Check if the file is readable
    sprintf(filename, "%s/%s", path, SO_NAME);
    fd = open(filename, O_RDONLY);
    if (fd >= 0) {
        strncpy(SO_NAME, filename, MAX_SO_PATH);
        close(fd);
    }

    // Clean up and exit
    free(filename);
    return (fd >= 0);
}

static int _getExecPath (char *path, unsigned long maxlen) {
#ifdef linux
    return readlink("/proc/self/exe", path, maxlen);
#endif

#ifdef DARWIN
    _NSGetExecutablePath(path, &maxlen);
    return maxlen;
#endif
}

static void _setSearchPath () {
    char  path[MAX_SO_PATH+1];
    int   count;
    char *p;

    /* Make sure that SO_NAME is not an absolute path. */
    if (SO_NAME[0] == '/')  return;

    /* Check the execution directory name. */
    memset(path, 0, sizeof(path));
    count = _getExecPath(path, MAX_SO_PATH);

    if (count > 0) {
        char *p = strrchr(path, '/');
        if (p == path)  ++p;
        if (p != 0)     *p = '\0';

        /* If there is a match, return immediately. */
        if (_checkPath(path))  return;
    }

    /* Check the current working directory. */
    p = getcwd(path, MAX_SO_PATH);
    if (p != 0)  _checkPath(path);
}

#endif


/*=========================================================================
| WINDOWS SUPPORT
 ========================================================================*/
#if defined(WIN32) || defined(_WIN32)

#include <stdio.h>
#include <windows.h>

#define DLL_HANDLE           HINSTANCE
#define dlopen(name, flags)  LoadLibraryA(name)
#define dlsym(handle, name)  GetProcAddress(handle, name)
#define dlerror()            "Exiting program"
#define SO_NAME              API_NAME ".dll"

/*
 * Use the default Windows DLL loading rules:
 *   1.  The directory from which the application binary was loaded.
 *   2.  The application's current directory.
 *   3a. [Windows NT/2000/XP only] 32-bit system directory
 *       (default: c:\winnt\System32)
 *   3b. 16-bit system directory
 *       (default: c:\winnt\System or c:\windows\system)
 *   4.  The windows directory
 *       (default: c:\winnt or c:\windows)
 *   5.  The directories listed in the PATH environment variable
 */
static void _setSearchPath () {
    /* Do nothing */
}

#endif


/*=========================================================================
| SHARED LIBRARY LOADER
 ========================================================================*/
/* The error conditions can be customized depending on the application. */
static void *_loadFunction (const char *name, int *result) {
    static DLL_HANDLE handle = 0;
    void * function = 0;

    /* Load the shared library if necessary */
    if (handle == 0) {
        u32 (*version) (void);
        u16 sw_version;
        u16 api_version_req;

        _setSearchPath();
        handle = dlopen(SO_NAME, RTLD_LAZY);
        if (handle == 0) {
#if API_DEBUG
            fprintf(stderr, "Unable to load %s\n", SO_NAME);
            fprintf(stderr, "%s\n", dlerror());
#endif
            *result = API_UNABLE_TO_LOAD_LIBRARY;
            return 0;
        }

        version = (void *)dlsym(handle, "pd_c_version");
        if (version == 0) {
#if API_DEBUG
            fprintf(stderr, "Unable to bind pd_c_version() in %s\n",
                    SO_NAME);
            fprintf(stderr, "%s\n", dlerror());
#endif
            handle  = 0;
            *result = API_INCOMPATIBLE_LIBRARY;
            return 0;
        }

        sw_version      = (u16)((version() >>  0) & 0xffff);
        api_version_req = (u16)((version() >> 16) & 0xffff);
        if (sw_version  < API_REQ_SW_VERSION ||
            API_HEADER_VERSION < api_version_req)
        {
#if API_DEBUG
            fprintf(stderr, "\nIncompatible versions:\n");

            fprintf(stderr, "  Header version  = v%d.%02d  ",
                    (API_HEADER_VERSION >> 8) & 0xff, API_HEADER_VERSION & 0xff);

            if (sw_version < API_REQ_SW_VERSION)
                fprintf(stderr, "(requires library >= %d.%02d)\n",
                        (API_REQ_SW_VERSION >> 8) & 0xff,
                        API_REQ_SW_VERSION & 0xff);
            else
                fprintf(stderr, "(library version OK)\n");


            fprintf(stderr, "  Library version = v%d.%02d  ",
                    (sw_version >> 8) & 0xff,
                    (sw_version >> 0) & 0xff);

            if (API_HEADER_VERSION < api_version_req)
                fprintf(stderr, "(requires header >= %d.%02d)\n",
                        (api_version_req >> 8) & 0xff,
                        (api_version_req >> 0) & 0xff);
            else
                fprintf(stderr, "(header version OK)\n");
#endif
            handle  = 0;
            *result = API_INCOMPATIBLE_LIBRARY;
            return 0;
        }
    }

    /* Bind the requested function in the shared library */
    function = (void *)dlsym(handle, name);
    *result  = function ? API_OK : API_UNABLE_TO_LOAD_FUNCTION;
    return function;
}


/*=========================================================================
| FUNCTIONS
 ========================================================================*/
static int (*c_pd_find_devices) (int, u16 *) = 0;
int pd_find_devices (
    int   num_devices,
    u16 * devices
)
{
    if (c_pd_find_devices == 0) {
        int res = 0;
        if (!(c_pd_find_devices = _loadFunction("c_pd_find_devices", &res)))
            return res;
    }
    return c_pd_find_devices(num_devices, devices);
}


static int (*c_pd_find_devices_ext) (int, u16 *, int, u32 *) = 0;
int pd_find_devices_ext (
    int   num_devices,
    u16 * devices,
    int   num_ids,
    u32 * unique_ids
)
{
    if (c_pd_find_devices_ext == 0) {
        int res = 0;
        if (!(c_pd_find_devices_ext = _loadFunction("c_pd_find_devices_ext", &res)))
            return res;
    }
    return c_pd_find_devices_ext(num_devices, devices, num_ids, unique_ids);
}


static Pd (*c_pd_open) (int) = 0;
Pd pd_open (
    int port_number
)
{
    if (c_pd_open == 0) {
        int res = 0;
        if (!(c_pd_open = _loadFunction("c_pd_open", &res)))
            return res;
    }
    return c_pd_open(port_number);
}


static Pd (*c_pd_open_ext) (int, PdExt *) = 0;
Pd pd_open_ext (
    int     port_number,
    PdExt * pd_ext
)
{
    if (c_pd_open_ext == 0) {
        int res = 0;
        if (!(c_pd_open_ext = _loadFunction("c_pd_open_ext", &res)))
            return res;
    }
    return c_pd_open_ext(port_number, pd_ext);
}


static int (*c_pd_close) (Pd) = 0;
int pd_close (
    Pd pd
)
{
    if (c_pd_close == 0) {
        int res = 0;
        if (!(c_pd_close = _loadFunction("c_pd_close", &res)))
            return res;
    }
    return c_pd_close(pd);
}


static int (*c_pd_port) (Pd) = 0;
int pd_port (
    Pd pd
)
{
    if (c_pd_port == 0) {
        int res = 0;
        if (!(c_pd_port = _loadFunction("c_pd_port", &res)))
            return res;
    }
    return c_pd_port(pd);
}


static u32 (*c_pd_unique_id) (Pd) = 0;
u32 pd_unique_id (
    Pd pd
)
{
    if (c_pd_unique_id == 0) {
        int res = 0;
        if (!(c_pd_unique_id = _loadFunction("c_pd_unique_id", &res)))
            return res;
    }
    return c_pd_unique_id(pd);
}


static const char * (*c_pd_status_string) (int) = 0;
const char * pd_status_string (
    int status
)
{
    if (c_pd_status_string == 0) {
        int res = 0;
        if (!(c_pd_status_string = _loadFunction("c_pd_status_string", &res)))
            return 0;
    }
    return c_pd_status_string(status);
}


static int (*c_pd_version) (Pd, PdVersion *) = 0;
int pd_version (
    Pd          pd,
    PdVersion * version
)
{
    if (c_pd_version == 0) {
        int res = 0;
        if (!(c_pd_version = _loadFunction("c_pd_version", &res)))
            return res;
    }
    return c_pd_version(pd, version);
}


static u32 (*c_pd_sleep_ms) (u32) = 0;
u32 pd_sleep_ms (
    u32 milliseconds
)
{
    if (c_pd_sleep_ms == 0) {
        int res = 0;
        if (!(c_pd_sleep_ms = _loadFunction("c_pd_sleep_ms", &res)))
            return res;
    }
    return c_pd_sleep_ms(milliseconds);
}


static int (*c_pd_capture_start) (Pd) = 0;
int pd_capture_start (
    Pd pd
)
{
    if (c_pd_capture_start == 0) {
        int res = 0;
        if (!(c_pd_capture_start = _loadFunction("c_pd_capture_start", &res)))
            return res;
    }
    return c_pd_capture_start(pd);
}


static int (*c_pd_capture_stop) (Pd) = 0;
int pd_capture_stop (
    Pd pd
)
{
    if (c_pd_capture_stop == 0) {
        int res = 0;
        if (!(c_pd_capture_stop = _loadFunction("c_pd_capture_stop", &res)))
            return res;
    }
    return c_pd_capture_stop(pd);
}


static int (*c_pd_usbpd_read_bits) (Pd, PdReadInfo *, u32 *, u32 *, u32, u08 *) = 0;
int pd_usbpd_read_bits (
    Pd           pd,
    PdReadInfo * read_info,
    u32 *        bits_length,
    u32 *        preamble_length,
    u32          max_bytes,
    u08 *        bits
)
{
    if (c_pd_usbpd_read_bits == 0) {
        int res = 0;
        if (!(c_pd_usbpd_read_bits = _loadFunction("c_pd_usbpd_read_bits", &res)))
            return res;
    }
    return c_pd_usbpd_read_bits(pd, read_info, bits_length, preamble_length, max_bytes, bits);
}


static int (*c_pd_usbpd_decode_bits) (Pd, u32, const u08 *, u32 *, u32 *, u32, u08 *) = 0;
int pd_usbpd_decode_bits (
    Pd          pd,
    u32         bytes_length,
    const u08 * bits,
    u32 *       header,
    u32 *       crc,
    u32         max_bytes,
    u08 *       data
)
{
    if (c_pd_usbpd_decode_bits == 0) {
        int res = 0;
        if (!(c_pd_usbpd_decode_bits = _loadFunction("c_pd_usbpd_decode_bits", &res)))
            return res;
    }
    return c_pd_usbpd_decode_bits(pd, bytes_length, bits, header, crc, max_bytes, data);
}


static int (*c_pd_usbpd_read_data) (Pd, PdReadInfo *, u32 *, u32 *, u32 *, u32, u08 *) = 0;
int pd_usbpd_read_data (
    Pd           pd,
    PdReadInfo * read_info,
    u32 *        preamble_length,
    u32 *        header,
    u32 *        crc,
    u32          max_bytes,
    u08 *        data
)
{
    if (c_pd_usbpd_read_data == 0) {
        int res = 0;
        if (!(c_pd_usbpd_read_data = _loadFunction("c_pd_usbpd_read_data", &res)))
            return res;
    }
    return c_pd_usbpd_read_data(pd, read_info, preamble_length, header, crc, max_bytes, data);
}


static int (*c_pd_usbpd_read_iv) (Pd, PdReadInfo *, int *) = 0;
int pd_usbpd_read_iv (
    Pd           pd,
    PdReadInfo * read_info,
    int *        val
)
{
    if (c_pd_usbpd_read_iv == 0) {
        int res = 0;
        if (!(c_pd_usbpd_read_iv = _loadFunction("c_pd_usbpd_read_iv", &res)))
            return res;
    }
    return c_pd_usbpd_read_iv(pd, read_info, val);
}


