/*=========================================================================
| Power Delivery Analyzer Library
|--------------------------------------------------------------------------
| Copyright (c) 2016-2024 Total Phase, Inc.
| All rights reserved.
| www.totalphase.com
|
| Redistribution and use of this file in source and binary forms, with
| or without modification, are permitted provided that the following
| conditions are met:
|
| - Redistributions of source code must retain the above copyright
|   notice, this list of conditions, and the following disclaimer.
|
| - Redistributions in binary form must reproduce the above copyright
|   notice, this list of conditions, and the following disclaimer in the
|   documentation or other materials provided with the distribution.
|
| - This file must only be used to interface with Total Phase products.
|   The names of Total Phase and its contributors must not be used to
|   endorse or promote products derived from this software.
|
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
| "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING BUT NOT
| LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
| FOR A PARTICULAR PURPOSE, ARE DISCLAIMED.  IN NO EVENT WILL THE
| COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
| INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING
| BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
| LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
| ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
| POSSIBILITY OF SUCH DAMAGE.
|--------------------------------------------------------------------------
| To access Total Phase Power Delivery analyzers through the API:
|
| 1) Use one of the following shared objects:
|      pd.so       --  Linux or macOS shared object
|      pd.dll      --  Windows dynamic link library
|
| 2) Along with one of the following language modules:
|      pd.c/h      --  C/C++ API header file and interface module
|      pd_py.py    --  Python API
|      pd.cs       --  C# .NET source
|      pd_net.dll  --  Compiled .NET binding
 ========================================================================*/


#ifndef __pd_h__
#define __pd_h__

#ifdef __cplusplus
extern "C" {
#endif


/*=========================================================================
| TYPEDEFS
 ========================================================================*/
#ifndef TOTALPHASE_DATA_TYPES
#define TOTALPHASE_DATA_TYPES

#ifndef _MSC_VER
/* C99-compliant compilers (GCC) */
#include <stdint.h>
typedef uint8_t   u08;
typedef uint16_t  u16;
typedef uint32_t  u32;
typedef uint64_t  u64;
typedef int8_t    s08;
typedef int16_t   s16;
typedef int32_t   s32;
typedef int64_t   s64;

#else
/* Microsoft compilers (Visual C++) */
typedef unsigned __int8   u08;
typedef unsigned __int16  u16;
typedef unsigned __int32  u32;
typedef unsigned __int64  u64;
typedef signed   __int8   s08;
typedef signed   __int16  s16;
typedef signed   __int32  s32;
typedef signed   __int64  s64;

#endif /* __MSC_VER */

typedef float   f32;
typedef double  f64;

#endif /* TOTALPHASE_DATA_TYPES */


/*=========================================================================
| DEBUG
 ========================================================================*/
/* Set the following macro to '1' for debugging */
#define PD_DEBUG 0


/*=========================================================================
| VERSION
 ========================================================================*/
#define PD_HEADER_VERSION  0x0200   /* v2.00 */


/*=========================================================================
| STATUS CODES
 ========================================================================*/
/*
 * All API functions return an integer which is the result of the
 * transaction, or a status code if negative.  The status codes are
 * defined as follows:
 */
enum PdStatus {
    /* General codes (0 to -99) */
    PD_OK                      =    0,
    PD_UNABLE_TO_LOAD_LIBRARY  =   -1,
    PD_UNABLE_TO_LOAD_DRIVER   =   -2,
    PD_UNABLE_TO_LOAD_FUNCTION =   -3,
    PD_INCOMPATIBLE_LIBRARY    =   -4,
    PD_INCOMPATIBLE_DEVICE     =   -5,
    PD_INCOMPATIBLE_DRIVER     =   -6,
    PD_COMMUNICATION_ERROR     =   -7,
    PD_UNABLE_TO_OPEN          =   -8,
    PD_UNABLE_TO_CLOSE         =   -9,
    PD_INVALID_HANDLE          =  -10,
    PD_CONFIG_ERROR            =  -11,
    PD_STILL_ACTIVE            =  -12,
    PD_FUNCTION_NOT_AVAILABLE  =  -13,

    PD_READ_EMPTY              = -100
};
#ifndef __cplusplus
typedef enum PdStatus PdStatus;
#endif


/*=========================================================================
| GENERAL TYPE DEFINITIONS
 ========================================================================*/
/* Pd handle type definition */
typedef int Pd;

/*
 * Pd version matrix.
 *
 * This matrix describes the various version dependencies
 * of Pd components.  It can be used to determine
 * which component caused an incompatibility error.
 *
 * All version numbers are of the format:
 *   (major << 8) | minor
 *
 * ex. v1.20 would be encoded as:  0x0114
 */
struct PdVersion {
    /* Software, firmware, and hardware versions. */
    u16 software;
    u16 firmware;
    u16 hardware;

    /* Firmware requires that software must be >= this version. */
    u16 sw_req_by_fw;

    /* Software requires that firmware must be >= this version. */
    u16 fw_req_by_sw;

    /* Software requires that the API interface must be >= this version. */
    u16 api_req_by_sw;
};
#ifndef __cplusplus
typedef struct PdVersion PdVersion;
#endif


/*=========================================================================
| GENERAL API
 ========================================================================*/
/*
 * Get a list of ports to which Pd devices are attached.
 *
 * num_devices = maximum number of elements to return
 * devices     = array into which the port numbers are returned
 *
 * Each element of the array is written with the port number.
 * Devices that are in-use are ORed with PD_PORT_NOT_FREE
 * (0x8000).
 *
 * ex.  devices are attached to ports 0, 1, 2
 *      ports 0 and 2 are available, and port 1 is in-use.
 *      array => 0x0000, 0x8001, 0x0002
 *
 * If the array is NULL, it is not filled with any values.
 * If there are more devices than the array size, only the
 * first nmemb port numbers will be written into the array.
 *
 * Returns the number of devices found, regardless of the
 * array size.
 */
#define PD_PORT_NOT_FREE 0x8000
int pd_find_devices (
    int   num_devices,
    u16 * devices
);


/*
 * Get a list of ports to which Pd devices are attached
 *
 * This function is the same as pd_find_devices() except that
 * it returns the unique IDs of each Pd device.  The IDs
 * are guaranteed to be non-zero if valid.
 *
 * The IDs are the unsigned integer representation of the 10-digit
 * serial numbers.
 */
int pd_find_devices_ext (
    int   num_devices,
    u16 * devices,
    int   num_ids,
    u32 * unique_ids
);


/*
 * Open the Pd port.
 *
 * The port number is a zero-indexed integer.
 *
 * The port number is the same as that obtained from the
 * pd_find_devices() function above.
 *
 * Returns an Pd handle, which is guaranteed to be
 * greater than zero if it is valid.
 *
 * This function is recommended for use in simple applications
 * where extended information is not required.  For more complex
 * applications, the use of pd_open_ext() is recommended.
 */
Pd pd_open (
    int port_number
);


/*
 * Open the Pd port, returning extended information
 * in the supplied structure.  Behavior is otherwise identical
 * to pd_open() above.  If 0 is passed as the pointer to the
 * structure, this function is exactly equivalent to pd_open().
 *
 * The structure is zeroed before the open is attempted.
 * It is filled with whatever information is available.
 *
 * For example, if the hardware version is not filled, then
 * the device could not be queried for its version number.
 *
 * This function is recommended for use in complex applications
 * where extended information is required.  For more simple
 * applications, the use of pd_open() is recommended.
 */
struct PdExt {
    /* Version matrix */
    PdVersion   version;

    /* Features of this device. */
    int         features;
};
#ifndef __cplusplus
typedef struct PdExt PdExt;
#endif

#define PD_FEATURE_NONE 0x00000000
#define PD_FEATURE_USBPD 0x00000001
#define PD_FEATURE_USBPD_IV 0x00000002
Pd pd_open_ext (
    int     port_number,
    PdExt * pd_ext
);


/* Close the Pd port. */
int pd_close (
    Pd pd
);


/*
 * Return the port for this Pd handle.
 *
 * The port number is a zero-indexed integer.
 */
int pd_port (
    Pd pd
);


/*
 * Return the unique ID for this Pd adapter.
 * IDs are guaranteed to be non-zero if valid.
 * The ID is the unsigned integer representation of the
 * 10-digit serial number.
 */
u32 pd_unique_id (
    Pd pd
);


/*
 * Return the status string for the given status code.
 * If the code is not valid or the library function cannot
 * be loaded, return a NULL string.
 */
const char * pd_status_string (
    int status
);


/*
 * Return the version matrix for the device attached to the
 * given handle.  If the handle is 0 or invalid, only the
 * software and required api versions are set.
 */
int pd_version (
    Pd          pd,
    PdVersion * version
);


/*
 * Sleep for the specified number of milliseconds
 * Accuracy depends on the operating system scheduler
 * Returns the number of milliseconds slept
 */
u32 pd_sleep_ms (
    u32 milliseconds
);



/*=========================================================================
| CAPTURE API
 ==========================================================================
| Start capturi*/
int pd_capture_start (
    Pd pd
);


/* Stop capturing data */
int pd_capture_stop (
    Pd pd
);



/*=========================================================================
| PD CC API
 ==========================================================================
| Valid with pd_usbpd_read_bits and pd_usbpd_read_da*/
#define PD_STATUS_USBPD_ERR_MASK 0x0000000f
#define PD_STATUS_USBPD_ERR_PREAMBLE 0x00000001
#define PD_STATUS_USBPD_ERR_SOP 0x00000002
#define PD_STATUS_USBPD_ERR_HEADER 0x00000003
#define PD_STATUS_USBPD_ERR_DATA 0x00000004
#define PD_STATUS_USBPD_ERR_CRC 0x00000005
#define PD_STATUS_USBPD_ERR_EOP 0x00000006
#define PD_STATUS_USBPD_ERR_BAD_CRC 0x01000000
#define PD_STATUS_USBPD_ERR_UNKNOWN_TYPE 0x02000000
/* Valid with pd_usbpd_read_bits and pd_usbpd_read_data */
#define PD_EVENT_USBPD_CC_MASK 0xf0000000
#define PD_EVENT_USBPD_CC_SHIFT 28
#define PD_EVENT_USBPD_CC1 0x00000000
#define PD_EVENT_USBPD_CC2 0x10000000
#define PD_EVENT_USBPD_POL_CHANGE 0x00000008
#define PD_EVENT_USBPD_SOP_MASK 0x00000007
#define PD_EVENT_USBPD_SOP 0x00000000
#define PD_EVENT_USBPD_SOP_PRIME 0x00000001
#define PD_EVENT_USBPD_SOP_DPRIME 0x00000002
#define PD_EVENT_USBPD_SOP_PRIME_DEBUG 0x00000003
#define PD_EVENT_USBPD_SOP_DPRIME_DEBUG 0x00000004
#define PD_EVENT_USBPD_HARD_RESET 0x00000006
#define PD_EVENT_USBPD_CABLE_RESET 0x00000007
#define PD_EVENT_USBPD_EXTENDED_HEADER 0x00000010
/* Valid only with pd_usbpd_read_iv */
#define PD_EVENT_USBPD_IV_SOURCE_MASK 0x0f000000
#define PD_EVENT_USBPD_IV_SHIFT 24
#define PD_EVENT_USBPD_IV_VBUS_VOLTAGE 0x00000000
#define PD_EVENT_USBPD_IV_VBUS_CURRENT 0x01000000
#define PD_EVENT_USBPD_IV_VCONN_VOLTAGE 0x02000000
#define PD_EVENT_USBPD_IV_VCONN_CURRENT 0x03000000
#define PD_EVENT_USBPD_IV_CC1_VOLTAGE 0x04000000
#define PD_EVENT_USBPD_IV_CC2_VOLTAGE 0x06000000
struct PdReadInfo {
    u64 timestamp;
    u64 duration;
    u32 status;
    u32 events;
};
#ifndef __cplusplus
typedef struct PdReadInfo PdReadInfo;
#endif

/* Read PD bits */
int pd_usbpd_read_bits (
    Pd           pd,
    PdReadInfo * read_info,
    u32 *        bits_length,
    u32 *        preamble_length,
    u32          max_bytes,
    u08 *        bits
);


/* Decode PD bits */
#define PD_HEADER_USBPD_STANDARD_MASK 0x0000ffff
#define PD_HEADER_USBPD_EXTENDED_MASK 0xffff0000
int pd_usbpd_decode_bits (
    Pd          pd,
    u32         bytes_length,
    const u08 * bits,
    u32 *       header,
    u32 *       crc,
    u32         max_bytes,
    u08 *       data
);


/* Read PD data return u08 */
int pd_usbpd_read_data (
    Pd           pd,
    PdReadInfo * read_info,
    u32 *        preamble_length,
    u32 *        header,
    u32 *        crc,
    u32          max_bytes,
    u08 *        data
);


/* Read IV data */
int pd_usbpd_read_iv (
    Pd           pd,
    PdReadInfo * read_info,
    int *        val
);




#ifdef __cplusplus
}
#endif

#endif  /* __pd_h__ */
