/*=========================================================================
| Power Delivery Analyzer Library
|--------------------------------------------------------------------------
| Copyright (c) 2016-2024 Total Phase, Inc.
| All rights reserved.
| www.totalphase.com
|
| Redistribution and use of this file in source and binary forms, with
| or without modification, are permitted provided that the following
| conditions are met:
|
| - Redistributions of source code must retain the above copyright
|   notice, this list of conditions, and the following disclaimer.
|
| - Redistributions in binary form must reproduce the above copyright
|   notice, this list of conditions, and the following disclaimer in the
|   documentation or other materials provided with the distribution.
|
| - This file must only be used to interface with Total Phase products.
|   The names of Total Phase and its contributors must not be used to
|   endorse or promote products derived from this software.
|
| THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
| "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING BUT NOT
| LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
| FOR A PARTICULAR PURPOSE, ARE DISCLAIMED.  IN NO EVENT WILL THE
| COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
| INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING
| BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
| LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
| CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
| LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
| ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
| POSSIBILITY OF SUCH DAMAGE.
|--------------------------------------------------------------------------
| To access Total Phase Power Delivery analyzers through the API:
|
| 1) Use one of the following shared objects:
|      pd.so       --  Linux or macOS shared object
|      pd.dll      --  Windows dynamic link library
|
| 2) Along with one of the following language modules:
|      pd.c/h      --  C/C++ API header file and interface module
|      pd_py.py    --  Python API
|      pd.cs       --  C# .NET source
|      pd_net.dll  --  Compiled .NET binding
 ========================================================================*/

using System;
using System.Reflection;
using System.Runtime.InteropServices;

[assembly: AssemblyTitleAttribute("Pd .NET binding")]
[assembly: AssemblyDescriptionAttribute(".NET binding for Pd")]
[assembly: AssemblyCompanyAttribute("Total Phase, Inc.")]
[assembly: AssemblyProductAttribute("Pd")]
[assembly: AssemblyCopyrightAttribute("Total Phase, Inc. 2024")]

namespace TotalPhase {

public enum PdStatus : int {
    /* General codes (0 to -99) */
    PD_OK                      =    0,
    PD_UNABLE_TO_LOAD_LIBRARY  =   -1,
    PD_UNABLE_TO_LOAD_DRIVER   =   -2,
    PD_UNABLE_TO_LOAD_FUNCTION =   -3,
    PD_INCOMPATIBLE_LIBRARY    =   -4,
    PD_INCOMPATIBLE_DEVICE     =   -5,
    PD_INCOMPATIBLE_DRIVER     =   -6,
    PD_COMMUNICATION_ERROR     =   -7,
    PD_UNABLE_TO_OPEN          =   -8,
    PD_UNABLE_TO_CLOSE         =   -9,
    PD_INVALID_HANDLE          =  -10,
    PD_CONFIG_ERROR            =  -11,
    PD_STILL_ACTIVE            =  -12,
    PD_FUNCTION_NOT_AVAILABLE  =  -13,

    PD_READ_EMPTY              = -100
}


public class PdApi {

/*=========================================================================
| HELPER FUNCTIONS / CLASSES
 ========================================================================*/
static long tp_min(long x, long y) { return x < y ? x : y; }

private class GCContext {
    GCHandle[] handles;
    int index;
    public GCContext () {
        handles = new GCHandle[16];
        index   = 0;
    }
    public void add (GCHandle gch) {
        handles[index] = gch;
        index++;
    }
    public void free () {
        while (index != 0) {
            index--;
            handles[index].Free();
        }
    }
}

/*=========================================================================
| VERSION
 ========================================================================*/
[DllImport ("pd")]
private static extern int pd_c_version ();

public const int PD_API_VERSION    = 0x0200;   // v2.00
public const int PD_REQ_SW_VERSION = 0x0128;   // v1.40

private static short PD_SW_VERSION;
private static short PD_REQ_API_VERSION;
private static bool  PD_LIBRARY_LOADED;

static PdApi () {
    PD_SW_VERSION      = (short)(pd_c_version() & 0xffff);
    PD_REQ_API_VERSION = (short)((pd_c_version() >> 16) & 0xffff);
    PD_LIBRARY_LOADED  = 
        ((PD_SW_VERSION >= PD_REQ_SW_VERSION) &&
         (PD_API_VERSION >= PD_REQ_API_VERSION));
}

/*=========================================================================
| STATUS CODES
 ========================================================================*/
/*
 * All API functions return an integer which is the result of the
 * transaction, or a status code if negative.  The status codes are
 * defined as follows:
 */
// enum PdStatus  (from declaration above)
//     PD_OK                      =    0
//     PD_UNABLE_TO_LOAD_LIBRARY  =   -1
//     PD_UNABLE_TO_LOAD_DRIVER   =   -2
//     PD_UNABLE_TO_LOAD_FUNCTION =   -3
//     PD_INCOMPATIBLE_LIBRARY    =   -4
//     PD_INCOMPATIBLE_DEVICE     =   -5
//     PD_INCOMPATIBLE_DRIVER     =   -6
//     PD_COMMUNICATION_ERROR     =   -7
//     PD_UNABLE_TO_OPEN          =   -8
//     PD_UNABLE_TO_CLOSE         =   -9
//     PD_INVALID_HANDLE          =  -10
//     PD_CONFIG_ERROR            =  -11
//     PD_STILL_ACTIVE            =  -12
//     PD_FUNCTION_NOT_AVAILABLE  =  -13
//     PD_READ_EMPTY              = -100


/*=========================================================================
| GENERAL TYPE DEFINITIONS
 ========================================================================*/
/* Pd handle type definition */
/* typedef Pd => int */

/*
 * Pd version matrix.
 *
 * This matrix describes the various version dependencies
 * of Pd components.  It can be used to determine
 * which component caused an incompatibility error.
 *
 * All version numbers are of the format:
 *   (major << 8) | minor
 *
 * ex. v1.20 would be encoded as:  0x0114
 */
[StructLayout(LayoutKind.Sequential)]
public struct PdVersion {
    /* Software, firmware, and hardware versions. */
    public ushort software;
    public ushort firmware;
    public ushort hardware;

    /* Firmware requires that software must be >= this version. */
    public ushort sw_req_by_fw;

    /* Software requires that firmware must be >= this version. */
    public ushort fw_req_by_sw;

    /* Software requires that the API interface must be >= this version. */
    public ushort api_req_by_sw;
}


/*=========================================================================
| GENERAL API
 ========================================================================*/
/*
 * Get a list of ports to which Pd devices are attached.
 *
 * num_devices = maximum number of elements to return
 * devices     = array into which the port numbers are returned
 *
 * Each element of the array is written with the port number.
 * Devices that are in-use are ORed with PD_PORT_NOT_FREE
 * (0x8000).
 *
 * ex.  devices are attached to ports 0, 1, 2
 *      ports 0 and 2 are available, and port 1 is in-use.
 *      array => 0x0000, 0x8001, 0x0002
 *
 * If the array is NULL, it is not filled with any values.
 * If there are more devices than the array size, only the
 * first nmemb port numbers will be written into the array.
 *
 * Returns the number of devices found, regardless of the
 * array size.
 */
public const ushort PD_PORT_NOT_FREE = 0x8000;
public static int pd_find_devices (
    int       num_devices,
    ushort[]  devices
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    int devices_num_devices = (int)tp_min(num_devices, devices.Length);
    return net_pd_find_devices(devices_num_devices, devices);
}

/*
 * Get a list of ports to which Pd devices are attached
 *
 * This function is the same as pd_find_devices() except that
 * it returns the unique IDs of each Pd device.  The IDs
 * are guaranteed to be non-zero if valid.
 *
 * The IDs are the unsigned integer representation of the 10-digit
 * serial numbers.
 */
public static int pd_find_devices_ext (
    int       num_devices,
    ushort[]  devices,
    int       num_ids,
    uint[]    unique_ids
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    int devices_num_devices = (int)tp_min(num_devices, devices.Length);
    int unique_ids_num_ids = (int)tp_min(num_ids, unique_ids.Length);
    return net_pd_find_devices_ext(devices_num_devices, devices, unique_ids_num_ids, unique_ids);
}

/*
 * Open the Pd port.
 *
 * The port number is a zero-indexed integer.
 *
 * The port number is the same as that obtained from the
 * pd_find_devices() function above.
 *
 * Returns an Pd handle, which is guaranteed to be
 * greater than zero if it is valid.
 *
 * This function is recommended for use in simple applications
 * where extended information is not required.  For more complex
 * applications, the use of pd_open_ext() is recommended.
 */
public static int pd_open (
    int  port_number
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    return net_pd_open(port_number);
}

/*
 * Open the Pd port, returning extended information
 * in the supplied structure.  Behavior is otherwise identical
 * to pd_open() above.  If 0 is passed as the pointer to the
 * structure, this function is exactly equivalent to pd_open().
 *
 * The structure is zeroed before the open is attempted.
 * It is filled with whatever information is available.
 *
 * For example, if the hardware version is not filled, then
 * the device could not be queried for its version number.
 *
 * This function is recommended for use in complex applications
 * where extended information is required.  For more simple
 * applications, the use of pd_open() is recommended.
 */
[StructLayout(LayoutKind.Sequential)]
public struct PdExt {
    /* Version matrix */
    public PdVersion version;

    /* Features of this device. */
    public int       features;
}

public const int PD_FEATURE_NONE = 0x00000000;
public const int PD_FEATURE_USBPD = 0x00000001;
public const int PD_FEATURE_USBPD_IV = 0x00000002;
public static int pd_open_ext (
    int        port_number,
    ref PdExt  pd_ext
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    return net_pd_open_ext(port_number, ref pd_ext);
}

/* Close the Pd port. */
public static int pd_close (
    int  pd
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    return net_pd_close(pd);
}

/*
 * Return the port for this Pd handle.
 *
 * The port number is a zero-indexed integer.
 */
public static int pd_port (
    int  pd
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    return net_pd_port(pd);
}

/*
 * Return the unique ID for this Pd adapter.
 * IDs are guaranteed to be non-zero if valid.
 * The ID is the unsigned integer representation of the
 * 10-digit serial number.
 */
public static uint pd_unique_id (
    int  pd
)
{
    if (!PD_LIBRARY_LOADED) return 0;
    return net_pd_unique_id(pd);
}

/*
 * Return the status string for the given status code.
 * If the code is not valid or the library function cannot
 * be loaded, return a NULL string.
 */
public static string pd_status_string (
    int  status
)
{
    if (!PD_LIBRARY_LOADED) return null;
    return Marshal.PtrToStringAnsi(net_pd_status_string(status));
}

/*
 * Return the version matrix for the device attached to the
 * given handle.  If the handle is 0 or invalid, only the
 * software and required api versions are set.
 */
public static int pd_version (
    int            pd,
    ref PdVersion  version
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    return net_pd_version(pd, ref version);
}

/*
 * Sleep for the specified number of milliseconds
 * Accuracy depends on the operating system scheduler
 * Returns the number of milliseconds slept
 */
public static uint pd_sleep_ms (
    uint  milliseconds
)
{
    if (!PD_LIBRARY_LOADED) return 0;
    return net_pd_sleep_ms(milliseconds);
}


/*=========================================================================
| CAPTURE API
 ==========================================================================
| Start capturi*/
public static int pd_capture_start (
    int  pd
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    return net_pd_capture_start(pd);
}

/* Stop capturing data */
public static int pd_capture_stop (
    int  pd
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    return net_pd_capture_stop(pd);
}


/*=========================================================================
| PD CC API
 ==========================================================================
| Valid with pd_usbpd_read_bits and pd_usbpd_read_da*/
public const uint PD_STATUS_USBPD_ERR_MASK = 0x0000000f;
public const uint PD_STATUS_USBPD_ERR_PREAMBLE = 0x00000001;
public const uint PD_STATUS_USBPD_ERR_SOP = 0x00000002;
public const uint PD_STATUS_USBPD_ERR_HEADER = 0x00000003;
public const uint PD_STATUS_USBPD_ERR_DATA = 0x00000004;
public const uint PD_STATUS_USBPD_ERR_CRC = 0x00000005;
public const uint PD_STATUS_USBPD_ERR_EOP = 0x00000006;
public const uint PD_STATUS_USBPD_ERR_BAD_CRC = 0x01000000;
public const uint PD_STATUS_USBPD_ERR_UNKNOWN_TYPE = 0x02000000;
/* Valid with pd_usbpd_read_bits and pd_usbpd_read_data */
public const uint PD_EVENT_USBPD_CC_MASK = 0xf0000000;
public const uint PD_EVENT_USBPD_CC_SHIFT = 28;
public const uint PD_EVENT_USBPD_CC1 = 0x00000000;
public const uint PD_EVENT_USBPD_CC2 = 0x10000000;
public const uint PD_EVENT_USBPD_POL_CHANGE = 0x00000008;
public const uint PD_EVENT_USBPD_SOP_MASK = 0x00000007;
public const uint PD_EVENT_USBPD_SOP = 0x00000000;
public const uint PD_EVENT_USBPD_SOP_PRIME = 0x00000001;
public const uint PD_EVENT_USBPD_SOP_DPRIME = 0x00000002;
public const uint PD_EVENT_USBPD_SOP_PRIME_DEBUG = 0x00000003;
public const uint PD_EVENT_USBPD_SOP_DPRIME_DEBUG = 0x00000004;
public const uint PD_EVENT_USBPD_HARD_RESET = 0x00000006;
public const uint PD_EVENT_USBPD_CABLE_RESET = 0x00000007;
public const uint PD_EVENT_USBPD_EXTENDED_HEADER = 0x00000010;
/* Valid only with pd_usbpd_read_iv */
public const uint PD_EVENT_USBPD_IV_SOURCE_MASK = 0x0f000000;
public const uint PD_EVENT_USBPD_IV_SHIFT = 24;
public const uint PD_EVENT_USBPD_IV_VBUS_VOLTAGE = 0x00000000;
public const uint PD_EVENT_USBPD_IV_VBUS_CURRENT = 0x01000000;
public const uint PD_EVENT_USBPD_IV_VCONN_VOLTAGE = 0x02000000;
public const uint PD_EVENT_USBPD_IV_VCONN_CURRENT = 0x03000000;
public const uint PD_EVENT_USBPD_IV_CC1_VOLTAGE = 0x04000000;
public const uint PD_EVENT_USBPD_IV_CC2_VOLTAGE = 0x06000000;
[StructLayout(LayoutKind.Sequential)]
public struct PdReadInfo {
    public ulong timestamp;
    public ulong duration;
    public uint  status;
    public uint  events;
}

/* Read PD bits */
public static int pd_usbpd_read_bits (
    int             pd,
    ref PdReadInfo  read_info,
    ref uint        bits_length,
    ref uint        preamble_length,
    uint            max_bytes,
    byte[]          bits
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    uint bits_max_bytes = (uint)tp_min(max_bytes, bits.Length);
    return net_pd_usbpd_read_bits(pd, ref read_info, ref bits_length, ref preamble_length, bits_max_bytes, bits);
}

/* Decode PD bits */
public const uint PD_HEADER_USBPD_STANDARD_MASK = 0x0000ffff;
public const uint PD_HEADER_USBPD_EXTENDED_MASK = 0xffff0000;
public static int pd_usbpd_decode_bits (
    int       pd,
    uint      bytes_length,
    byte[]    bits,
    ref uint  header,
    ref uint  crc,
    uint      max_bytes,
    byte[]    data
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    uint bits_bytes_length = (uint)tp_min(bytes_length, bits.Length);
    uint data_max_bytes = (uint)tp_min(max_bytes, data.Length);
    return net_pd_usbpd_decode_bits(pd, bits_bytes_length, bits, ref header, ref crc, data_max_bytes, data);
}

/* Read PD data return u08 */
public static int pd_usbpd_read_data (
    int             pd,
    ref PdReadInfo  read_info,
    ref uint        preamble_length,
    ref uint        header,
    ref uint        crc,
    uint            max_bytes,
    byte[]          data
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    uint data_max_bytes = (uint)tp_min(max_bytes, data.Length);
    return net_pd_usbpd_read_data(pd, ref read_info, ref preamble_length, ref header, ref crc, data_max_bytes, data);
}

/* Read IV data */
public static int pd_usbpd_read_iv (
    int             pd,
    ref PdReadInfo  read_info,
    ref int         val
)
{
    if (!PD_LIBRARY_LOADED) return (int)PdStatus.PD_INCOMPATIBLE_LIBRARY;
    return net_pd_usbpd_read_iv(pd, ref read_info, ref val);
}


/*=========================================================================
| NATIVE DLL BINDINGS
 ========================================================================*/
[DllImport ("pd")]
private static extern int net_pd_find_devices (int num_devices, [Out] ushort[] devices);

[DllImport ("pd")]
private static extern int net_pd_find_devices_ext (int num_devices, [Out] ushort[] devices, int num_ids, [Out] uint[] unique_ids);

[DllImport ("pd")]
private static extern int net_pd_open (int port_number);

[DllImport ("pd")]
private static extern int net_pd_open_ext (int port_number, ref PdExt pd_ext);

[DllImport ("pd")]
private static extern int net_pd_close (int pd);

[DllImport ("pd")]
private static extern int net_pd_port (int pd);

[DllImport ("pd")]
private static extern uint net_pd_unique_id (int pd);

[DllImport ("pd")]
private static extern IntPtr net_pd_status_string (int status);

[DllImport ("pd")]
private static extern int net_pd_version (int pd, ref PdVersion version);

[DllImport ("pd")]
private static extern uint net_pd_sleep_ms (uint milliseconds);

[DllImport ("pd")]
private static extern int net_pd_capture_start (int pd);

[DllImport ("pd")]
private static extern int net_pd_capture_stop (int pd);

[DllImport ("pd")]
private static extern int net_pd_usbpd_read_bits (int pd, ref PdReadInfo read_info, ref uint bits_length, ref uint preamble_length, uint max_bytes, [Out] byte[] bits);

[DllImport ("pd")]
private static extern int net_pd_usbpd_decode_bits (int pd, uint bytes_length, [In] byte[] bits, ref uint header, ref uint crc, uint max_bytes, [Out] byte[] data);

[DllImport ("pd")]
private static extern int net_pd_usbpd_read_data (int pd, ref PdReadInfo read_info, ref uint preamble_length, ref uint header, ref uint crc, uint max_bytes, [Out] byte[] data);

[DllImport ("pd")]
private static extern int net_pd_usbpd_read_iv (int pd, ref PdReadInfo read_info, ref int val);


} // class PdApi

} // namespace TotalPhase
