Attribute VB_Name = "PD"
'==========================================================================
' Power Delivery Analyzer Library
'--------------------------------------------------------------------------
' Copyright (c) 2016-2024 Total Phase, Inc.
' All rights reserved.
' www.totalphase.com
'
' Redistribution and use of this file in source and binary forms, with
' or without modification, are permitted provided that the following
' conditions are met:
'
' - Redistributions of source code must retain the above copyright
'   notice, this list of conditions, and the following disclaimer.
'
' - Redistributions in binary form must reproduce the above copyright
'   notice, this list of conditions, and the following disclaimer in the
'   documentation or other materials provided with the distribution.
'
' - This file must only be used to interface with Total Phase products.
'   The names of Total Phase and its contributors must not be used to
'   endorse or promote products derived from this software.
'
' THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
' "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING BUT NOT
' LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
' FOR A PARTICULAR PURPOSE, ARE DISCLAIMED.  IN NO EVENT WILL THE
' COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
' INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING
' BUT NOT LIMITED TO PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
' LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
' CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
' LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
' ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
' POSSIBILITY OF SUCH DAMAGE.
'--------------------------------------------------------------------------
' To access Total Phase Power Delivery analyzers through the API:
'
' 1) Use one of the following shared objects:
'      pd.so       --  Linux or macOS shared object
'      pd.dll      --  Windows dynamic link library
'
' 2) Along with one of the following language modules:
'      pd.c/h      --  C/C++ API header file and interface module
'      pd_py.py    --  Python API
'      pd.cs       --  C# .NET source
'      pd_net.dll  --  Compiled .NET binding
'==========================================================================

' Data types:
'    Handles are Long
'
'    Fixed bit-width types
'      8-bit => Byte
'     16-bit => Integer
'     32-bit => Long
'     64-bit => LongLong

Public Type LongLong
    LoValue As Long
    HiValue As Long
End Type

Public Enum PdStatus
    ' General codes (0 to -99)
    PD_OK                      =    0
    PD_UNABLE_TO_LOAD_LIBRARY  =   -1
    PD_UNABLE_TO_LOAD_DRIVER   =   -2
    PD_UNABLE_TO_LOAD_FUNCTION =   -3
    PD_INCOMPATIBLE_LIBRARY    =   -4
    PD_INCOMPATIBLE_DEVICE     =   -5
    PD_INCOMPATIBLE_DRIVER     =   -6
    PD_COMMUNICATION_ERROR     =   -7
    PD_UNABLE_TO_OPEN          =   -8
    PD_UNABLE_TO_CLOSE         =   -9
    PD_INVALID_HANDLE          =  -10
    PD_CONFIG_ERROR            =  -11
    PD_STILL_ACTIVE            =  -12
    PD_FUNCTION_NOT_AVAILABLE  =  -13

    PD_READ_EMPTY              = -100
End Enum

Public Type PdVersion
    ' Software, firmware, and hardware versions.
    software        as Integer
    firmware        as Integer
    hardware        as Integer

    ' Firmware requires that software must be >= this version.
    sw_req_by_fw    as Integer

    ' Software requires that firmware must be >= this version.
    fw_req_by_sw    as Integer

    ' Software requires that the API interface must be >= this version.
    api_req_by_sw   as Integer
End Type

Public Const PD_PORT_NOT_FREE = &H8000

Public Type PdExt
    ' Version matrix
    version    as PdVersion

    ' Features of this device.
    features   as Long
End Type

Public Const PD_FEATURE_NONE = &H00000000
Public Const PD_FEATURE_USBPD = &H00000001
Public Const PD_FEATURE_USBPD_IV = &H00000002

Public Const PD_STATUS_USBPD_ERR_MASK = &H0000000f
Public Const PD_STATUS_USBPD_ERR_PREAMBLE = &H00000001
Public Const PD_STATUS_USBPD_ERR_SOP = &H00000002
Public Const PD_STATUS_USBPD_ERR_HEADER = &H00000003
Public Const PD_STATUS_USBPD_ERR_DATA = &H00000004
Public Const PD_STATUS_USBPD_ERR_CRC = &H00000005
Public Const PD_STATUS_USBPD_ERR_EOP = &H00000006
Public Const PD_STATUS_USBPD_ERR_BAD_CRC = &H01000000
Public Const PD_STATUS_USBPD_ERR_UNKNOWN_TYPE = &H02000000

Public Const PD_EVENT_USBPD_CC_MASK = &Hf0000000
Public Const PD_EVENT_USBPD_CC_SHIFT = 28
Public Const PD_EVENT_USBPD_CC1 = &H00000000
Public Const PD_EVENT_USBPD_CC2 = &H10000000
Public Const PD_EVENT_USBPD_POL_CHANGE = &H00000008
Public Const PD_EVENT_USBPD_SOP_MASK = &H00000007
Public Const PD_EVENT_USBPD_SOP = &H00000000
Public Const PD_EVENT_USBPD_SOP_PRIME = &H00000001
Public Const PD_EVENT_USBPD_SOP_DPRIME = &H00000002
Public Const PD_EVENT_USBPD_SOP_PRIME_DEBUG = &H00000003
Public Const PD_EVENT_USBPD_SOP_DPRIME_DEBUG = &H00000004
Public Const PD_EVENT_USBPD_HARD_RESET = &H00000006
Public Const PD_EVENT_USBPD_CABLE_RESET = &H00000007
Public Const PD_EVENT_USBPD_EXTENDED_HEADER = &H00000010

Public Const PD_EVENT_USBPD_IV_SOURCE_MASK = &H0f000000
Public Const PD_EVENT_USBPD_IV_SHIFT = 24
Public Const PD_EVENT_USBPD_IV_VBUS_VOLTAGE = &H00000000
Public Const PD_EVENT_USBPD_IV_VBUS_CURRENT = &H01000000
Public Const PD_EVENT_USBPD_IV_VCONN_VOLTAGE = &H02000000
Public Const PD_EVENT_USBPD_IV_VCONN_CURRENT = &H03000000
Public Const PD_EVENT_USBPD_IV_CC1_VOLTAGE = &H04000000
Public Const PD_EVENT_USBPD_IV_CC2_VOLTAGE = &H06000000

Public Type PdReadInfo
    timestamp   as LongLong
    duration    as LongLong
    status      as Long
    events      as Long
End Type

Public Const PD_HEADER_USBPD_STANDARD_MASK = &H0000ffff
Public Const PD_HEADER_USBPD_EXTENDED_MASK = &Hffff0000

Private Declare Function vb6_pd_find_devices Lib "pd.dll" (ByVal num_devices As Long, ByRef devices As Integer) as Long
Private Declare Function vb6_pd_find_devices_ext Lib "pd.dll" (ByVal num_devices As Long, ByRef devices As Integer, ByVal num_ids As Long, ByRef unique_ids As Long) as Long
Private Declare Function vb6_pd_open Lib "pd.dll" (ByVal port_number As Long) as Long
Private Declare Function vb6_pd_open_ext Lib "pd.dll" (ByVal port_number As Long, ByRef pd_ext As PdExt) as Long
Private Declare Function vb6_pd_close Lib "pd.dll" (ByVal pd As Long) as Long
Private Declare Function vb6_pd_port Lib "pd.dll" (ByVal pd As Long) as Long
Private Declare Function vb6_pd_unique_id Lib "pd.dll" (ByVal pd As Long) as Long
Private Declare Function vb6_pd_status_string Lib "pd.dll" (ByVal status As Long) as String
Private Declare Function vb6_pd_version Lib "pd.dll" (ByVal pd As Long, ByRef version As PdVersion) as Long
Private Declare Function vb6_pd_sleep_ms Lib "pd.dll" (ByVal milliseconds As Long) as Long
Private Declare Function vb6_pd_capture_start Lib "pd.dll" (ByVal pd As Long) as Long
Private Declare Function vb6_pd_capture_stop Lib "pd.dll" (ByVal pd As Long) as Long
Private Declare Function vb6_pd_usbpd_read_bits Lib "pd.dll" (ByVal pd As Long, ByRef read_info As PdReadInfo, ByRef bits_length As Long, ByRef preamble_length As Long, ByVal max_bytes As Long, ByRef bits As Byte) as Long
Private Declare Function vb6_pd_usbpd_decode_bits Lib "pd.dll" (ByVal pd As Long, ByVal bytes_length As Long, ByRef bits As Byte, ByRef header As Long, ByRef crc As Long, ByVal max_bytes As Long, ByRef data As Byte) as Long
Private Declare Function vb6_pd_usbpd_read_data Lib "pd.dll" (ByVal pd As Long, ByRef read_info As PdReadInfo, ByRef preamble_length As Long, ByRef header As Long, ByRef crc As Long, ByVal max_bytes As Long, ByRef data As Byte) as Long
Private Declare Function vb6_pd_usbpd_read_iv Lib "pd.dll" (ByVal pd As Long, ByRef read_info As PdReadInfo, ByRef val As Long) as Long


'==========================================================================
' VERSION
'==========================================================================
Public Const PD_API_VERSION    = &H0200   ' v2.00
Public Const PD_REQ_SW_VERSION = &H0128   ' v1.40

Private Declare Function pd_std_version Lib "pd.dll" () as Long
Dim PD_VERSION_OK as Boolean

Private Function check_version () as Boolean
    If PD_VERSION_OK Then
        check_version = True
    Else
        Dim sw_version As Long
        Dim req_api_version As Long
        sw_version = pd_std_version() And &Hffff
        req_api_version = (pd_std_version() / 65536) And &Hffff
        PD_VERSION_OK = ((sw_version >= PD_REQ_SW_VERSION) And (PD_API_VERSION >= req_api_version))
        check_version = PD_VERSION_OK
    End If
End Function

'==========================================================================
' HELPER FUNCTIONS
'==========================================================================
Public Function tp_min (ByVal x As Variant, ByVal y As Variant) as Variant
    If x > y Then
        tp_min = y
    Else
        tp_min = x
    End If
End Function


'==========================================================================
' STATUS CODES
'==========================================================================
' All API functions return an integer which is the result of the
' transaction, or a status code if negative.  The status codes are
' defined as follows:

'==========================================================================
' GENERAL TYPE DEFINITIONS
'==========================================================================
' Pd handle type definition
' typedef Pd => Long

' Pd version matrix.
'
' This matrix describes the various version dependencies
' of Pd components.  It can be used to determine
' which component caused an incompatibility error.
'
' All version numbers are of the format:
'   (major << 8) | minor
'
' ex. v1.20 would be encoded as:  0x0114

'==========================================================================
' GENERAL API
'==========================================================================
' Get a list of ports to which Pd devices are attached.
'
' num_devices = maximum number of elements to return
' devices     = array into which the port numbers are returned
'
' Each element of the array is written with the port number.
' Devices that are in-use are ORed with PD_PORT_NOT_FREE
' (0x8000).
'
' ex.  devices are attached to ports 0, 1, 2
'      ports 0 and 2 are available, and port 1 is in-use.
'      array => 0x0000, 0x8001, 0x0002
'
' If the array is NULL, it is not filled with any values.
' If there are more devices than the array size, only the
' first nmemb port numbers will be written into the array.
'
' Returns the number of devices found, regardless of the
' array size.

Public Function pd_find_devices (ByVal num_devices As Long, ByRef devices() As Integer) as Long
    If check_version() Then
        Dim devices_num_devices As Long
        devices_num_devices = tp_min(num_devices, UBound(devices) - LBound(devices) + 1)
        pd_find_devices = vb6_pd_find_devices(devices_num_devices, devices(0))
    Else
        pd_find_devices = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Get a list of ports to which Pd devices are attached
'
' This function is the same as pd_find_devices() except that
' it returns the unique IDs of each Pd device.  The IDs
' are guaranteed to be non-zero if valid.
'
' The IDs are the unsigned integer representation of the 10-digit
' serial numbers.

Public Function pd_find_devices_ext (ByVal num_devices As Long, ByRef devices() As Integer, ByVal num_ids As Long, ByRef unique_ids() As Long) as Long
    If check_version() Then
        Dim devices_num_devices As Long
        devices_num_devices = tp_min(num_devices, UBound(devices) - LBound(devices) + 1)
        Dim unique_ids_num_ids As Long
        unique_ids_num_ids = tp_min(num_ids, UBound(unique_ids) - LBound(unique_ids) + 1)
        pd_find_devices_ext = vb6_pd_find_devices_ext(devices_num_devices, devices(0), unique_ids_num_ids, unique_ids(0))
    Else
        pd_find_devices_ext = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Open the Pd port.
'
' The port number is a zero-indexed integer.
'
' The port number is the same as that obtained from the
' pd_find_devices() function above.
'
' Returns an Pd handle, which is guaranteed to be
' greater than zero if it is valid.
'
' This function is recommended for use in simple applications
' where extended information is not required.  For more complex
' applications, the use of pd_open_ext() is recommended.

Public Function pd_open (ByVal port_number As Long) as Long
    If check_version() Then
        pd_open = vb6_pd_open(port_number)
    Else
        pd_open = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Open the Pd port, returning extended information
' in the supplied structure.  Behavior is otherwise identical
' to pd_open() above.  If 0 is passed as the pointer to the
' structure, this function is exactly equivalent to pd_open().
'
' The structure is zeroed before the open is attempted.
' It is filled with whatever information is available.
'
' For example, if the hardware version is not filled, then
' the device could not be queried for its version number.
'
' This function is recommended for use in complex applications
' where extended information is required.  For more simple
' applications, the use of pd_open() is recommended.

Public Function pd_open_ext (ByVal port_number As Long, ByRef pd_ext As PdExt) as Long
    If check_version() Then
        pd_open_ext = vb6_pd_open_ext(port_number, pd_ext)
    Else
        pd_open_ext = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Close the Pd port.

Public Function pd_close (ByVal pd As Long) as Long
    If check_version() Then
        pd_close = vb6_pd_close(pd)
    Else
        pd_close = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Return the port for this Pd handle.
'
' The port number is a zero-indexed integer.

Public Function pd_port (ByVal pd As Long) as Long
    If check_version() Then
        pd_port = vb6_pd_port(pd)
    Else
        pd_port = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Return the unique ID for this Pd adapter.
' IDs are guaranteed to be non-zero if valid.
' The ID is the unsigned integer representation of the
' 10-digit serial number.

Public Function pd_unique_id (ByVal pd As Long) as Long
    If check_version() Then
        pd_unique_id = vb6_pd_unique_id(pd)
    Else
        pd_unique_id = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Return the status string for the given status code.
' If the code is not valid or the library function cannot
' be loaded, return a NULL string.

Public Function pd_status_string (ByVal status As Long) as String
    If check_version() Then
        pd_status_string = vb6_pd_status_string(status)
    Else
        pd_status_string = vbNullString
    End If
End Function


' Return the version matrix for the device attached to the
' given handle.  If the handle is 0 or invalid, only the
' software and required api versions are set.

Public Function pd_version (ByVal pd As Long, ByRef version As PdVersion) as Long
    If check_version() Then
        pd_version = vb6_pd_version(pd, version)
    Else
        pd_version = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Sleep for the specified number of milliseconds
' Accuracy depends on the operating system scheduler
' Returns the number of milliseconds slept

Public Function pd_sleep_ms (ByVal milliseconds As Long) as Long
    If check_version() Then
        pd_sleep_ms = vb6_pd_sleep_ms(milliseconds)
    Else
        pd_sleep_ms = PD_INCOMPATIBLE_LIBRARY
    End If
End Function



'==========================================================================
' CAPTURE API
'==========================================================================
' Start capturing

Public Function pd_capture_start (ByVal pd As Long) as Long
    If check_version() Then
        pd_capture_start = vb6_pd_capture_start(pd)
    Else
        pd_capture_start = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Stop capturing data

Public Function pd_capture_stop (ByVal pd As Long) as Long
    If check_version() Then
        pd_capture_stop = vb6_pd_capture_stop(pd)
    Else
        pd_capture_stop = PD_INCOMPATIBLE_LIBRARY
    End If
End Function



'==========================================================================
' PD CC API
'==========================================================================
' Valid with pd_usbpd_read_bits and pd_usbpd_read_data
' Valid with pd_usbpd_read_bits and pd_usbpd_read_data
' Valid only with pd_usbpd_read_iv
' Read PD bits

Public Function pd_usbpd_read_bits (ByVal pd As Long, ByRef read_info As PdReadInfo, ByRef bits_length As Long, ByRef preamble_length As Long, ByVal max_bytes As Long, ByRef bits() As Byte) as Long
    If check_version() Then
        Dim bits_max_bytes As Long
        bits_max_bytes = tp_min(max_bytes, UBound(bits) - LBound(bits) + 1)
        pd_usbpd_read_bits = vb6_pd_usbpd_read_bits(pd, read_info, bits_length, preamble_length, bits_max_bytes, bits(0))
    Else
        pd_usbpd_read_bits = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Decode PD bits

Public Function pd_usbpd_decode_bits (ByVal pd As Long, ByVal bytes_length As Long, ByRef bits() As Byte, ByRef header As Long, ByRef crc As Long, ByVal max_bytes As Long, ByRef data() As Byte) as Long
    If check_version() Then
        Dim bits_bytes_length As Long
        bits_bytes_length = tp_min(bytes_length, UBound(bits) - LBound(bits) + 1)
        Dim data_max_bytes As Long
        data_max_bytes = tp_min(max_bytes, UBound(data) - LBound(data) + 1)
        pd_usbpd_decode_bits = vb6_pd_usbpd_decode_bits(pd, bits_bytes_length, bits(0), header, crc, data_max_bytes, data(0))
    Else
        pd_usbpd_decode_bits = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Read PD data return u08

Public Function pd_usbpd_read_data (ByVal pd As Long, ByRef read_info As PdReadInfo, ByRef preamble_length As Long, ByRef header As Long, ByRef crc As Long, ByVal max_bytes As Long, ByRef data() As Byte) as Long
    If check_version() Then
        Dim data_max_bytes As Long
        data_max_bytes = tp_min(max_bytes, UBound(data) - LBound(data) + 1)
        pd_usbpd_read_data = vb6_pd_usbpd_read_data(pd, read_info, preamble_length, header, crc, data_max_bytes, data(0))
    Else
        pd_usbpd_read_data = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


' Read IV data

Public Function pd_usbpd_read_iv (ByVal pd As Long, ByRef read_info As PdReadInfo, ByRef val As Long) as Long
    If check_version() Then
        pd_usbpd_read_iv = vb6_pd_usbpd_read_iv(pd, read_info, val)
    Else
        pd_usbpd_read_iv = PD_INCOMPATIBLE_LIBRARY
    End If
End Function


